from stats.stats import IterableStats
from nfstream import NFPlugin
import numpy as np # for bytes distribution

class ResReqDiffTime(NFPlugin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def on_init(self, packet, flow):
        '''
        on_init(self, packet, flow): Method called at flow creation.
        '''
        flow.udps.req_res_time_diff = list() 
        flow.udps.current_flow_direction = 0 # 0 for forward, 1 for backward
        flow.udps.current_flow_direction_timestamp = packet.time
        flow.udps.min_req_res_time_diff = 0
        flow.udps.max_req_res_time_diff = 0
        flow.udps.mean_req_res_time_diff = 0
        flow.udps.stddev_req_res_time_diff = 0
        flow.udps.variance_req_res_time_diff = 0
        flow.udps.coeff_of_var_req_res_time_diff = 0
        flow.udps.skew_from_median_req_res_time_diff = 0

    def on_update(self, packet, flow):
        if packet.direction != flow.udps.current_flow_direction:
            flow.udps.req_res_time_diff.append(packet.time - flow.udps.current_flow_direction_timestamp)
            flow.udps.current_flow_direction = packet.direction
            flow.udps.current_flow_direction_timestamp = packet.time

    def on_expire(self, flow): 
        stats = IterableStats(flow.udps.req_res_time_diff)
        flow.udps.min_req_res_time_diff = stats.min()
        flow.udps.max_req_res_time_diff = stats.max()
        flow.udps.mean_req_res_time_diff = stats.average()
        flow.udps.median_req_res_time_diff = stats.median()
        flow.udps.stddev_req_res_time_diff = stats.std_deviation()
        flow.udps.variance_req_res_time_diff = stats.variance()
        flow.udps.coeff_of_var_req_res_time_diff = stats.coeff_of_variation()
        flow.udps.skew_from_median_req_res_time_diff = stats.skew_from_median()
        
        # Cleanup
        del flow.udps.current_flow_direction_timestamp
        del flow.udps.current_flow_direction






from setuptools import setup, find_packages
'''
python setup.py sdist
twine upload dist/*
'''

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='OSF_EIMTC',
    version='0.1.11',
    description='A Framework for Encrypted Internet and Malicious Traffic Classification.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/neyney10/PCAPFeatureExtractor',
    keywords='nfstream, pcap, network, deep-learning, extraction',
    license='MIT',
    author="Ariel University",
    author_email='ofek.bader@msmail.ariel.ac.il',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={
        'EIMTC': ['tools/*'],
    },
    install_requires=[
        'scikit-learn',
        'NFStream',
        'pandas',
        'numpy',
        'scapy',
        'pyasn',
        'click',
    ],
)


import numpy as np
from runstats import *
from nfstream import NFPlugin




class Clump_Flow(NFPlugin):
    '''
    Clumps:
        Description:
            The plugin 'clumps' the packets of each flow by the direction src2dst or dst2src,
            for each clump in flow clump starts and end when the next packet change direction or last flow packet.
            The clumps then extracted stastical information from the clumps groups.
        INFO:
            The paper we used as refrence to the clumps is "The Challenge of Only One Flow Problem for Traffic Classification in Identity Obfuscation Environments":
            In the paper the authors call the clumps as 'subflows' such forward (src2dst) subflows and backward (dst2src) subflows.
            We used only the subflow features:
                -PACKET NUMBER (SIZE)
                -PACKET BYTES (LEN)
                -BYTES PER PACKET
            In addition we added few more statistical features to each one of the features above.

            the authors of the paper:
            HONG-YEN CHEN & TSUNG-NAN LIN

        Coded by Adi Lichy,
        Edited by Ofek Bader,
        Ariel University.
    '''

    def on_init(self, packet, flow):
        flow.udps.src2dst_clumps = Clumps()
        flow.udps.dst2src_clumps = Clumps()
        flow.udps.clumps = Clumps()
        flow.udps.direction = None
        flow.udps.clump = None
        self.on_update(packet,flow)

    def on_update(self, packet, flow):
        if flow.udps.direction is None or flow.udps.direction != packet.direction:#first packet direction or change of direction
            if flow.udps.clump is not None:#not the first clump
                if flow.udps.direction == 0:
                    flow.udps.clump.update_bytes_per_packet()
                    flow.udps.src2dst_clumps.add_clump(flow.udps.clump)
                    flow.udps.src2dst_clumps.update(flow.udps.clump.size,flow.udps.clump.len,flow.udps.clump.bytes_per_packet,flow.udps.clump.interarrival_time)
                    flow.udps.clumps.add_clump(flow.udps.clump)
                    flow.udps.clumps.update(flow.udps.clump.size,flow.udps.clump.len,flow.udps.clump.bytes_per_packet,flow.udps.clump.interarrival_time)
                else:
                    flow.udps.clump.update_bytes_per_packet()
                    flow.udps.dst2src_clumps.add_clump(flow.udps.clump)
                    flow.udps.dst2src_clumps.update(flow.udps.clump.size,flow.udps.clump.len,flow.udps.clump.bytes_per_packet,flow.udps.clump.interarrival_time)
                    flow.udps.clumps.add_clump(flow.udps.clump)
                    flow.udps.clumps.update(flow.udps.clump.size,flow.udps.clump.len,flow.udps.clump.bytes_per_packet,flow.udps.clump.interarrival_time)
            flow.udps.direction = packet.direction
            flow.udps.clump = Clump(flow.udps.direction)
            flow.udps.clump.update_size(1)
            flow.udps.clump.update_len(packet.raw_size)
            flow.udps.clump.update_clump_time_stats(packet.time)
            flow.udps.clump.update_time(packet.delta_time)
        else:
            flow.udps.clump.update_size(1)
            flow.udps.clump.update_clump_time_stats(packet.time)
            flow.udps.clump.update_len(packet.raw_size)

    def on_expire(self, flow):
        #end of flow closing last clump
        if flow.udps.direction == 0:
            flow.udps.clump.update_bytes_per_packet()
            flow.udps.src2dst_clumps.add_clump(flow.udps.clump)
            flow.udps.src2dst_clumps.update(flow.udps.clump.size,flow.udps.clump.len,flow.udps.clump.bytes_per_packet,flow.udps.clump.interarrival_time)
            flow.udps.clumps.add_clump(flow.udps.clump)
            flow.udps.clumps.update(flow.udps.clump.size,flow.udps.clump.len,flow.udps.clump.bytes_per_packet,flow.udps.clump.interarrival_time)
        else:
            flow.udps.clump.update_bytes_per_packet()
            flow.udps.dst2src_clumps.add_clump(flow.udps.clump)
            flow.udps.dst2src_clumps.update(flow.udps.clump.size,flow.udps.clump.len,flow.udps.clump.bytes_per_packet,flow.udps.clump.interarrival_time)
            flow.udps.clumps.add_clump(flow.udps.clump)
            flow.udps.clumps.update(flow.udps.clump.size,flow.udps.clump.len,flow.udps.clump.bytes_per_packet,flow.udps.clump.interarrival_time)
        #gathering clumps results
        len_stats_src2dst = flow.udps.src2dst_clumps.get_len_stats()
        len_stats_dst2src = flow.udps.dst2src_clumps.get_len_stats()
        len_stats = flow.udps.clumps.get_len_stats()
        size_stats_src2dst = flow.udps.src2dst_clumps.get_size_stats()
        size_stats_dst2src = flow.udps.dst2src_clumps.get_size_stats()
        size_stats = flow.udps.clumps.get_size_stats()
        bytes_per_packet_stats_src2dst = flow.udps.src2dst_clumps.get_bytes_per_packet_stats()
        bytes_per_packet_stats_dst2src = flow.udps.dst2src_clumps.get_bytes_per_packet_stats()
        bytes_per_packet_stats = flow.udps.clumps.get_bytes_per_packet_stats()
        interarrival_time_stats_src2dst = flow.udps.src2dst_clumps.get_time_stats()
        interarrival_time_stats_dst2src = flow.udps.dst2src_clumps.get_time_stats()
        interarrival_time_stats = flow.udps.clumps.get_time_stats()
        
        #gather clump results
        src2dst_clump = flow.udps.src2dst_clumps.get_clumps()
        dst2src_clump = flow.udps.dst2src_clumps.get_clumps()
        clump = flow.udps.clumps.get_clumps()
        len_stats_src2dst_clump_max = []
        len_stats_src2dst_clump_min = []
        len_stats_src2dst_clump_mean = []
        len_stats_src2dst_clump_stddev = []
        len_stats_src2dst_clump_skewness = []
        len_stats_src2dst_clump_variance = []
        len_stats_src2dst_clump_kurtosis = []
        len_stats_dst2src_clump_max = []
        len_stats_dst2src_clump_min = []
        len_stats_dst2src_clump_mean = []
        len_stats_dst2src_clump_stddev = []
        len_stats_dst2src_clump_skewness = []
        len_stats_dst2src_clump_variance = []
        len_stats_dst2src_clump_kurtosis = []
        len_stats_clump_max = []
        len_stats_clump_min = []
        len_stats_clump_mean = []
        len_stats_clump_stddev = []
        len_stats_clump_skewness = []
        len_stats_clump_variance = []
        len_stats_clump_kurtosis = []

        interarrival_time_stats_src2dst_clump_max = []
        interarrival_time_stats_src2dst_clump_min = []
        interarrival_time_stats_src2dst_clump_mean = []
        interarrival_time_stats_src2dst_clump_stddev = []
        interarrival_time_stats_src2dst_clump_skewness = []
        interarrival_time_stats_src2dst_clump_variance = []
        interarrival_time_stats_src2dst_clump_kurtosis = []
        interarrival_time_stats_dst2src_clump_max = []
        interarrival_time_stats_dst2src_clump_min = []
        interarrival_time_stats_dst2src_clump_mean = []
        interarrival_time_stats_dst2src_clump_stddev = []
        interarrival_time_stats_dst2src_clump_skewness = []
        interarrival_time_stats_dst2src_clump_variance = []
        interarrival_time_stats_dst2src_clump_kurtosis = []
        interarrival_time_stats_clump_max = []
        interarrival_time_stats_clump_min = []
        interarrival_time_stats_clump_mean = []
        interarrival_time_stats_clump_stddev = []
        interarrival_time_stats_clump_skewness = []
        interarrival_time_stats_clump_variance = []
        interarrival_time_stats_clump_kurtosis = []


        for clump_number in range(len(clump)):
            if clump_number < len(src2dst_clump):
                current_len = src2dst_clump[clump_number].get_len_stats()
                current_time = src2dst_clump[clump_number].get_time_stats()
                #len
                len_stats_src2dst_clump_max.append(current_len['maximum'])
                len_stats_src2dst_clump_min.append(current_len['minimum'])
                len_stats_src2dst_clump_mean.append(current_len['mean'])
                len_stats_src2dst_clump_stddev.append(current_len['stddev'])
                len_stats_src2dst_clump_skewness.append(current_len['skewness'])
                len_stats_src2dst_clump_variance.append(current_len['variance'])
                len_stats_src2dst_clump_kurtosis.append(current_len['kurtosis'])
                #time
                interarrival_time_stats_src2dst_clump_max.append(current_time['maximum'])
                interarrival_time_stats_src2dst_clump_min.append(current_time['minimum'])
                interarrival_time_stats_src2dst_clump_mean.append(current_time['mean'])
                interarrival_time_stats_src2dst_clump_stddev.append(current_time['stddev'])
                interarrival_time_stats_src2dst_clump_skewness.append(current_time['skewness'])
                interarrival_time_stats_src2dst_clump_variance.append(current_time['variance'])
                interarrival_time_stats_src2dst_clump_kurtosis.append(current_time['kurtosis'])
            if clump_number < len(dst2src_clump):
                current_len = dst2src_clump[clump_number].get_len_stats()
                current_time = dst2src_clump[clump_number].get_time_stats()
                #len
                len_stats_dst2src_clump_max.append(current_len['maximum'])
                len_stats_dst2src_clump_min.append(current_len['minimum'])
                len_stats_dst2src_clump_mean.append(current_len['mean'])
                len_stats_dst2src_clump_stddev.append(current_len['stddev'])
                len_stats_dst2src_clump_skewness.append(current_len['skewness'])
                len_stats_dst2src_clump_variance.append(current_len['variance'])
                len_stats_dst2src_clump_kurtosis.append(current_len['kurtosis'])
                #time
                interarrival_time_stats_dst2src_clump_max.append(current_time['maximum'])
                interarrival_time_stats_dst2src_clump_min.append(current_time['minimum'])
                interarrival_time_stats_dst2src_clump_mean.append(current_time['mean'])
                interarrival_time_stats_dst2src_clump_stddev.append(current_time['stddev'])
                interarrival_time_stats_dst2src_clump_skewness.append(current_time['skewness'])
                interarrival_time_stats_dst2src_clump_variance.append(current_time['variance'])
                interarrival_time_stats_dst2src_clump_kurtosis.append(current_time['kurtosis'])

            current_len = clump[clump_number].get_len_stats()
            current_time = clump[clump_number].get_time_stats()
            #len
            len_stats_clump_max.append(current_len['maximum'])
            len_stats_clump_min.append(current_len['minimum'])
            len_stats_clump_mean.append(current_len['mean'])
            len_stats_clump_stddev.append(current_len['stddev'])
            len_stats_clump_skewness.append(current_len['skewness'])
            len_stats_clump_variance.append(current_len['variance'])
            len_stats_clump_kurtosis.append(current_len['kurtosis'])
            #time
            interarrival_time_stats_clump_max.append(current_time['maximum'])
            interarrival_time_stats_clump_min.append(current_time['minimum'])
            interarrival_time_stats_clump_mean.append(current_time['mean'])
            interarrival_time_stats_clump_stddev.append(current_time['stddev'])
            interarrival_time_stats_clump_skewness.append(current_time['skewness'])
            interarrival_time_stats_clump_variance.append(current_time['variance'])
            interarrival_time_stats_clump_kurtosis.append(current_time['kurtosis'])

        #src2dst
        #len
        flow.udps.src2dst_max_clumps_len =      len_stats_src2dst['maximum']
        flow.udps.src2dst_min_clumps_len =      len_stats_src2dst['minimum']
        flow.udps.src2dst_mean_clumps_len =     len_stats_src2dst['mean']
        flow.udps.src2dst_stddev_clumps_len =   len_stats_src2dst['stddev']
        flow.udps.src2dst_skewness_clumps_len = len_stats_src2dst['skewness']
        flow.udps.src2dst_variance_clumps_len = len_stats_src2dst['variance']
        flow.udps.src2dst_kurtosis_clumps_len = len_stats_src2dst['kurtosis']

        #size
        flow.udps.src2dst_max_clumps_size =      size_stats_src2dst['maximum']
        flow.udps.src2dst_min_clumps_size =      size_stats_src2dst['minimum']
        flow.udps.src2dst_mean_clumps_size =     size_stats_src2dst['mean']
        flow.udps.src2dst_stddev_clumps_size =   size_stats_src2dst['stddev']
        flow.udps.src2dst_skewness_clumps_size = size_stats_src2dst['skewness']
        flow.udps.src2dst_variance_clumps_size = size_stats_src2dst['variance']
        flow.udps.src2dst_kurtosis_clumps_size = size_stats_src2dst['kurtosis']

        #bytes_per_packet
        flow.udps.src2dst_max_clumps_bytes_per_packet =       bytes_per_packet_stats_src2dst['maximum']
        flow.udps.src2dst_min_clumps_bytes_per_packet =       bytes_per_packet_stats_src2dst['minimum']
        flow.udps.src2dst_mean_clumps_bytes_per_packet =      bytes_per_packet_stats_src2dst['mean']
        flow.udps.src2dst_stddev_clumps_bytes_per_packet =    bytes_per_packet_stats_src2dst['stddev']
        flow.udps.src2dst_skewness_clumps_bytes_per_packet =  bytes_per_packet_stats_src2dst['skewness']
        flow.udps.src2dst_variance_clumps_bytes_per_packet =  bytes_per_packet_stats_src2dst['variance']
        flow.udps.src2dst_kurtosis_clumps_bytes_per_packet =  bytes_per_packet_stats_src2dst['kurtosis']

        #interarrival_time
        flow.udps.src2dst_max_clumps_interarrival_time =       interarrival_time_stats_src2dst['maximum']
        flow.udps.src2dst_min_clumps_interarrival_time =       interarrival_time_stats_src2dst['minimum']
        flow.udps.src2dst_mean_clumps_interarrival_time =      interarrival_time_stats_src2dst['mean']
        flow.udps.src2dst_stddev_clumps_interarrival_time =    interarrival_time_stats_src2dst['stddev']
        flow.udps.src2dst_skewness_clumps_interarrival_time =  interarrival_time_stats_src2dst['skewness']
        flow.udps.src2dst_variance_clumps_interarrival_time =  interarrival_time_stats_src2dst['variance']
        flow.udps.src2dst_kurtosis_clumps_interarrival_time =  interarrival_time_stats_src2dst['kurtosis']

        #dst2src
        #len
        flow.udps.dst2src_max_clumps_len =      len_stats_dst2src['maximum']
        flow.udps.dst2src_min_clumps_len =      len_stats_dst2src['minimum']
        flow.udps.dst2src_mean_clumps_len =     len_stats_dst2src['mean']
        flow.udps.dst2src_stddev_clumps_len =   len_stats_dst2src['stddev']
        flow.udps.dst2src_skewness_clumps_len = len_stats_dst2src['skewness']
        flow.udps.dst2src_variance_clumps_len = len_stats_dst2src['variance']
        flow.udps.dst2src_kurtosis_clumps_len = len_stats_dst2src['kurtosis']

        #size
        flow.udps.dst2src_max_clumps_size =      size_stats_dst2src['maximum']
        flow.udps.dst2src_min_clumps_size =      size_stats_dst2src['minimum']
        flow.udps.dst2src_mean_clumps_size =     size_stats_dst2src['mean']
        flow.udps.dst2src_stddev_clumps_size =   size_stats_dst2src['stddev']
        flow.udps.dst2src_skewness_clumps_size = size_stats_dst2src['skewness']
        flow.udps.dst2src_variance_clumps_size = size_stats_dst2src['variance']
        flow.udps.dst2src_kurtosis_clumps_size = size_stats_dst2src['kurtosis']

        #bytes_per_packet
        flow.udps.dst2src_max_clumps_bytes_per_packet =       bytes_per_packet_stats_dst2src['maximum']
        flow.udps.dst2src_min_clumps_bytes_per_packet =       bytes_per_packet_stats_dst2src['minimum']
        flow.udps.dst2src_mean_clumps_bytes_per_packet =      bytes_per_packet_stats_dst2src['mean']
        flow.udps.dst2src_stddev_clumps_bytes_per_packet =    bytes_per_packet_stats_dst2src['stddev']
        flow.udps.dst2src_skewness_clumps_bytes_per_packet =  bytes_per_packet_stats_dst2src['skewness']
        flow.udps.dst2src_variance_clumps_bytes_per_packet =  bytes_per_packet_stats_dst2src['variance']
        flow.udps.dst2src_kurtosis_clumps_bytes_per_packet =  bytes_per_packet_stats_dst2src['kurtosis']

        #interarrival_time
        flow.udps.dst2src_max_clumps_interarrival_time =       interarrival_time_stats_dst2src['maximum']
        flow.udps.dst2src_min_clumps_interarrival_time =       interarrival_time_stats_dst2src['minimum']
        flow.udps.dst2src_mean_clumps_interarrival_time =      interarrival_time_stats_dst2src['mean']
        flow.udps.dst2src_stddev_clumps_interarrival_time =    interarrival_time_stats_dst2src['stddev']
        flow.udps.dst2src_skewness_clumps_interarrival_time =  interarrival_time_stats_dst2src['skewness']
        flow.udps.dst2src_variance_clumps_interarrival_time =  interarrival_time_stats_dst2src['variance']
        flow.udps.dst2src_kurtosis_clumps_interarrival_time =  interarrival_time_stats_dst2src['kurtosis']

        #bidirectional clumps
        #len
        flow.udps.max_clumps_len =      len_stats['maximum']
        flow.udps.min_clumps_len =      len_stats['minimum']
        flow.udps.mean_clumps_len =     len_stats['mean']
        flow.udps.stddev_clumps_len =   len_stats['stddev']
        flow.udps.skewness_clumps_len = len_stats['skewness']
        flow.udps.variance_clumps_len = len_stats['variance']
        flow.udps.kurtosis_clumps_len = len_stats['kurtosis']

        #size
        flow.udps.max_clumps_size =      size_stats['maximum']
        flow.udps.min_clumps_size =      size_stats['minimum']
        flow.udps.mean_clumps_size =     size_stats['mean']
        flow.udps.stddev_clumps_size =   size_stats['stddev']
        flow.udps.skewness_clumps_size = size_stats['skewness']
        flow.udps.variance_clumps_size = size_stats['variance']
        flow.udps.kurtosis_clumps_size = size_stats['kurtosis']

        #bytes_per_packet
        flow.udps.max_clumps_bytes_per_packet =       bytes_per_packet_stats['maximum']
        flow.udps.min_clumps_bytes_per_packet =       bytes_per_packet_stats['minimum']
        flow.udps.mean_clumps_bytes_per_packet =      bytes_per_packet_stats['mean']
        flow.udps.stddev_clumps_bytes_per_packet =    bytes_per_packet_stats['stddev']
        flow.udps.skewness_clumps_bytes_per_packet =  bytes_per_packet_stats['skewness']
        flow.udps.variance_clumps_bytes_per_packet =  bytes_per_packet_stats['variance']
        flow.udps.kurtosis_clumps_bytes_per_packet =  bytes_per_packet_stats['kurtosis']

        #interarrival_time
        flow.udps.max_clumps_interarrival_time =       interarrival_time_stats['maximum']
        flow.udps.min_clumps_interarrival_time =       interarrival_time_stats['minimum']
        flow.udps.mean_clumps_interarrival_time =      interarrival_time_stats['mean']
        flow.udps.stddev_clumps_interarrival_time =    interarrival_time_stats['stddev']
        flow.udps.skewness_clumps_interarrival_time =  interarrival_time_stats['skewness']
        flow.udps.variance_clumps_interarrival_time =  interarrival_time_stats['variance']
        flow.udps.kurtosis_clumps_interarrival_time =  interarrival_time_stats['kurtosis']
        
        #Clump
        #src2dst
        #len
        flow.udps.src2dst_max_clump_len =       len_stats_src2dst_clump_max
        flow.udps.src2dst_min_clump_len =       len_stats_src2dst_clump_min
        flow.udps.src2dst_mean_clump_len =      len_stats_src2dst_clump_mean 
        flow.udps.src2dst_stddev_clump_len =    len_stats_src2dst_clump_stddev
        flow.udps.src2dst_skewness_clump_len =  len_stats_src2dst_clump_skewness
        flow.udps.src2dst_variance_clump_len =  len_stats_src2dst_clump_variance
        flow.udps.src2dst_kurtosis_clump_len =  len_stats_src2dst_clump_kurtosis

        #interarrival_time
        flow.udps.src2dst_max_clump_interarrival_time =       interarrival_time_stats_src2dst_clump_max
        flow.udps.src2dst_min_clump_interarrival_time =       interarrival_time_stats_src2dst_clump_min
        flow.udps.src2dst_mean_clump_interarrival_time =      interarrival_time_stats_src2dst_clump_mean 
        flow.udps.src2dst_stddev_clump_interarrival_time =    interarrival_time_stats_src2dst_clump_stddev
        flow.udps.src2dst_skewness_clump_interarrival_time =  interarrival_time_stats_src2dst_clump_skewness
        flow.udps.src2dst_variance_clump_interarrival_time =  interarrival_time_stats_src2dst_clump_variance
        flow.udps.src2dst_kurtosis_clump_interarrival_time =  interarrival_time_stats_src2dst_clump_kurtosis


        #dst2src
        #len
        flow.udps.dst2src_max_clump_len =       len_stats_dst2src_clump_max
        flow.udps.dst2src_min_clump_len =       len_stats_dst2src_clump_min
        flow.udps.dst2src_mean_clump_len =      len_stats_dst2src_clump_mean 
        flow.udps.dst2src_stddev_clump_len =    len_stats_dst2src_clump_stddev
        flow.udps.dst2src_skewness_clump_len =  len_stats_dst2src_clump_skewness
        flow.udps.dst2src_variance_clump_len =  len_stats_dst2src_clump_variance
        flow.udps.dst2src_kurtosis_clump_len =  len_stats_dst2src_clump_kurtosis

        #interarrival_time
        flow.udps.dst2src_max_clump_interarrival_time =       interarrival_time_stats_dst2src_clump_max
        flow.udps.dst2src_min_clump_interarrival_time =       interarrival_time_stats_dst2src_clump_min
        flow.udps.dst2src_mean_clump_interarrival_time =      interarrival_time_stats_dst2src_clump_mean 
        flow.udps.dst2src_stddev_clump_interarrival_time =    interarrival_time_stats_dst2src_clump_stddev
        flow.udps.dst2src_skewness_clump_interarrival_time =  interarrival_time_stats_dst2src_clump_skewness
        flow.udps.dst2src_variance_clump_interarrival_time =  interarrival_time_stats_dst2src_clump_variance
        flow.udps.dst2src_kurtosis_clump_interarrival_time =  interarrival_time_stats_dst2src_clump_kurtosis

        #bidirectional
        #len
        flow.udps.max_clump_len =       len_stats_clump_max
        flow.udps.min_clump_len =       len_stats_clump_min
        flow.udps.mean_clump_len =      len_stats_clump_mean 
        flow.udps.stddev_clump_len =    len_stats_clump_stddev
        flow.udps.skewness_clump_len =  len_stats_clump_skewness
        flow.udps.variance_clump_len =  len_stats_clump_variance
        flow.udps.kurtosis_clump_len =  len_stats_clump_kurtosis

        #interarrival_time
        flow.udps.max_clump_interarrival_time =       interarrival_time_stats_clump_max
        flow.udps.min_clump_interarrival_time =       interarrival_time_stats_clump_min
        flow.udps.mean_clump_interarrival_time =      interarrival_time_stats_clump_mean 
        flow.udps.stddev_clump_interarrival_time =    interarrival_time_stats_clump_stddev
        flow.udps.skewness_clump_interarrival_time =  interarrival_time_stats_clump_skewness
        flow.udps.variance_clump_interarrival_time =  interarrival_time_stats_clump_variance
        flow.udps.kurtosis_clump_interarrival_time =  interarrival_time_stats_clump_kurtosis

        #delete variables
        del flow.udps.direction
        del flow.udps.clump
        del flow.udps.src2dst_clumps
        del flow.udps.dst2src_clumps
        del flow.udps.clumps



#Clump class
class Clump:

    def __init__(self,direction):
        self.direction = direction
        self.size = 0
        self.len = 0
        self.len_stats = Statistics()
        self.bytes_per_packet = 0
        self.interarrival_time = 0
        self.interarrival_time_stats = Statistics()
        self.previous_time = 0

    def get_time_stats(self):
        stats = {}
        stats['maximum'] = self.interarrival_time_stats.maximum() if self.interarrival_time_stats._count > 0 else 0
        stats['minimum'] = self.interarrival_time_stats.minimum() if self.interarrival_time_stats._count > 0 else 0
        stats['mean'] = self.interarrival_time_stats.mean()
        stats['stddev'] = self.interarrival_time_stats.stddev() if self.interarrival_time_stats._count >= 2 else 0
        stats['skewness'] = self.interarrival_time_stats.skewness() if stats['stddev'] != 0 else 0
        stats['variance'] = self.interarrival_time_stats.variance() if self.interarrival_time_stats._count >= 2 else 0
        stats['kurtosis'] = self.interarrival_time_stats.kurtosis() if stats['stddev'] != 0 else 0
        return stats
    
    def get_len_stats(self):
        stats = {}
        stats['maximum'] = self.len_stats.maximum() if self.len_stats._count > 0 else 0
        stats['minimum'] = self.len_stats.minimum() if self.len_stats._count > 0 else 0
        stats['mean'] = self.len_stats.mean()
        stats['stddev'] = self.len_stats.stddev() if self.len_stats._count >= 2 else 0
        stats['skewness'] = self.len_stats.skewness() if stats['stddev'] != 0 else 0
        stats['variance'] = self.len_stats.variance() if self.len_stats._count >= 2 else 0
        stats['kurtosis'] = self.len_stats.kurtosis() if stats['stddev'] != 0 else 0
        return stats


    def update(self,size,time,len,bytes_per_packet):
        self.update_size(size)
        self.update_time(time)
        self.update_len(len)
        self.update_bytes_per_packet(bytes_per_packet)

    def update_size(self,value):
        self.size += value

    def update_time(self,value):
        self.interarrival_time = value

    def update_clump_time_stats(self,value):
        if self.previous_time == 0:
            self.previous_time = value
            value = 0.0
        else:
            temp = value - self.previous_time
            self.previous_time = value
            value = temp
        self.interarrival_time_stats.push(value)

    def update_len(self,value):
        self.len_stats.push(value)
        self.len += value

    def update_bytes_per_packet(self):
        self.bytes_per_packet = self.len_stats.mean()
    
    def get_size(self):
        return self.size

    def get_time(self):
        return self.interarrival_time

    def get_len(self):
        return self.len

    def get_bytes_per_packet(self):
        return self.bytes_per_packet

    #Clump class END

#Clumps class
class Clumps:
    
    def __init__(self):
        self.clumps = []
        self.size = Statistics() # number of packets
        self.len = Statistics() # bytes
        self.bytes_per_packet = Statistics()
        self.interarrival_time = Statistics()# interarrival_time

    def add_clump(self,clump):
        self.clumps.append(clump)

    def update(self,size,len,bytes_per_packet,time):
        self.update_size(size)
        self.update_len(len)
        self.update_bytes_per_packet(bytes_per_packet)
        self.update_time(time)

    def update_size(self,value):
        self.size.push(value)

    def update_time(self,value):
        self.interarrival_time.push(value)

    def update_len(self,value):
        self.len.push(value)

    def update_bytes_per_packet(self,value):
        self.bytes_per_packet.push(value)

    def get_clumps(self):
        return self.clumps

    def get_size_stats(self):
        stats = {}
        stats['maximum'] = self.size.maximum() if self.size._count > 0 else 0
        stats['minimum'] = self.size.minimum() if self.size._count > 0 else 0
        stats['mean'] = self.size.mean()
        stats['stddev'] = self.size.stddev() if self.size._count >= 2 else 0
        stats['skewness'] = self.size.skewness() if stats['stddev'] != 0 else 0
        stats['variance'] = self.size.variance() if self.size._count >= 2 else 0
        stats['kurtosis'] = self.size.kurtosis() if stats['stddev'] != 0 else 0
        return stats

  
    def get_time_stats(self):
        stats = {}
        stats['maximum'] = self.interarrival_time.maximum() if self.interarrival_time._count > 0 else 0
        stats['minimum'] = self.interarrival_time.minimum() if self.interarrival_time._count > 0 else 0
        stats['mean'] = self.interarrival_time.mean()
        stats['stddev'] = self.interarrival_time.stddev() if self.interarrival_time._count >= 2 else 0
        stats['skewness'] = self.interarrival_time.skewness() if stats['stddev'] != 0 else 0
        stats['variance'] = self.interarrival_time.variance() if self.interarrival_time._count >= 2 else 0
        stats['kurtosis'] = self.interarrival_time.kurtosis() if stats['stddev'] != 0 else 0
        return stats

    def get_len_stats(self):
        stats = {}
        stats['maximum'] = self.len.maximum() if self.len._count > 0 else 0
        stats['minimum'] = self.len.minimum() if self.len._count > 0 else 0
        stats['mean'] = self.len.mean()
        stats['stddev'] = self.len.stddev() if self.len._count >= 2 else 0
        stats['skewness'] = self.len.skewness() if stats['stddev'] != 0 else 0
        stats['variance'] = self.len.variance() if self.len._count >= 2 else 0
        stats['kurtosis'] = self.len.kurtosis() if stats['stddev'] != 0 else 0
        return stats

    def get_bytes_per_packet_stats(self):
        stats = {}
        stats['maximum'] = self.bytes_per_packet.maximum() if self.bytes_per_packet._count > 0 else 0
        stats['minimum'] = self.bytes_per_packet.minimum() if self.bytes_per_packet._count > 0 else 0
        stats['mean'] = self.bytes_per_packet.mean()
        stats['stddev'] = self.bytes_per_packet.stddev() if self.bytes_per_packet._count >= 2 else 0
        stats['skewness'] = self.bytes_per_packet.skewness() if stats['stddev'] != 0 else 0
        stats['variance'] = self.bytes_per_packet.variance() if self.bytes_per_packet._count >= 2 else 0
        stats['kurtosis'] = self.bytes_per_packet.kurtosis() if stats['stddev'] != 0 else 0
        return stats
#Clumps class END


from nfstream import NFPlugin
from PIL import Image
import numpy as np
import os
from collections import namedtuple

class FlowPic2(NFPlugin):
    '''
    File name structure:
        flowpic-[flow_start_time_ms]-[src_ip]-[src_port]-[dst_ip]-[dst_port]-[protocol]-[direction].npz
    
    '''
    def __init__(self, save_path, 
                 time_per_subflow=30, 
                 min_time_per_subflow=None, 
                 min_packets_per_subflow=10, 
                 image_dims=(1500, 1500), 
                 MTU=1500):
        self.save_path = save_path
        self.time_per_subflow = time_per_subflow
        self.min_time_per_subflow = int(np.ceil(time_per_subflow - time_per_subflow*0.1666667)) if min_time_per_subflow == None else min_time_per_subflow
        self.min_packets_per_subflow = min_packets_per_subflow
        self.image_dims = image_dims
        self.mtu = MTU

    def on_init(self, packet, flow):
        flow.udps.src2dst_flowpic_data = _FlowPicInnerPlugin(flow.src2dst_first_seen_ms, self.mtu)
        flow.udps.dst2src_flowpic_data = None
        
        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        if packet.ip_size == 0 or packet.ip_size > self.image_dims[0]:
            # discard
            return

        if packet.direction == 0: # src -> dst
            if self.is_subflow_ended(flow.udps.src2dst_flowpic_data.start_time, packet.time, self.time_per_subflow):
                if flow.udps.dst2src_flowpic_data.count > self.min_packets_per_subflow:
                    self.save_histogram(flow.udps.src2dst_flowpic_data.on_expire(), 
                                        flow.udps.src2dst_flowpic_data.start_time, 
                                        packet.direction,
                                        flow)
                flow.udps.src2dst_flowpic_data = _FlowPicInnerPlugin(packet.time, self.mtu)
            flow.udps.src2dst_flowpic_data.on_update(packet)
        else:
            if flow.udps.dst2src_flowpic_data == None:
                flow.udps.dst2src_flowpic_data = _FlowPicInnerPlugin(flow.dst2src_first_seen_ms, self.mtu)
            if self.is_subflow_ended(flow.udps.dst2src_flowpic_data.start_time, packet.time, self.time_per_subflow):
                if flow.udps.dst2src_flowpic_data.count > self.min_packets_per_subflow:
                    self.save_histogram(flow.udps.dst2src_flowpic_data.on_expire(), 
                                        flow.udps.dst2src_flowpic_data.start_time, 
                                        packet.direction,
                                    flow)
                flow.udps.dst2src_flowpic_data = _FlowPicInnerPlugin(packet.time, self.mtu)
            flow.udps.dst2src_flowpic_data.on_update(packet)
        
    def on_expire(self, flow):
        ## SRC -> DST
        if flow.src2dst_packets > 0 \
            and flow.udps.src2dst_flowpic_data.count > self.min_packets_per_subflow \
            and self.is_subflow_ended(flow.udps.src2dst_flowpic_data.start_time, flow.src2dst_last_seen_ms, self.min_time_per_subflow):
            self.save_histogram(flow.udps.src2dst_flowpic_data.on_expire(), 
                                flow.udps.src2dst_flowpic_data.start_time,
                                0,
                                flow)
        ## DST -> SRC
        if flow.dst2src_packets > 0 \
            and flow.udps.dst2src_flowpic_data.count > self.min_packets_per_subflow \
            and self.is_subflow_ended(flow.udps.dst2src_flowpic_data.start_time, flow.dst2src_last_seen_ms, self.min_time_per_subflow):
            self.save_histogram(flow.udps.dst2src_flowpic_data.on_expire(), 
                                flow.udps.dst2src_flowpic_data.start_time,
                                1,
                                flow)
            
        # cleanup
        del flow.udps.src2dst_flowpic_data
        del flow.udps.dst2src_flowpic_data
 
    def save_histogram(self, hist, subflow_start_time, direction, flow):
        np.savez_compressed(
            os.path.join(
                self.save_path,
                '-'.join(['flowpic',
                    str(subflow_start_time), 
                    flow.src_ip, 
                    str(flow.src_port), 
                    flow.dst_ip, 
                    str(flow.dst_port), 
                    str(flow.protocol), 
                    'src2dst' if direction==0 else 'dst2src']
                )
            ), 
            flowpic=hist
        )
 
    def is_subflow_ended(self, start_time_ms, current_time_ms, time_per_subflow_s):
        return (current_time_ms - start_time_ms)/1000 > time_per_subflow_s


class _FlowPicInnerPlugin:
    '''
    NFPlugin-Like class to handle multiple subflows (slices of flows by time)
    For example, 130s flow, with slices of 60s, has 3 subflows.
    
    Note: Unidirectional.
    '''
    def __init__(self, subflow_start_time, mtu) -> None:
        self.start_time = subflow_start_time
        self.sizes = []
        self.timestamps = []
        self.mtu = mtu
        self.count = 0
    
    def on_update(self, packet):
        self.count += 1
        relative_time_ms = packet.time \
                        - self.start_time
        relative_time_s = relative_time_ms / 1000
        self.sizes.append(packet.ip_size)
        self.timestamps.append(relative_time_s)
        
    def on_expire(self):
        return self._flow_2d_histogram(self.sizes, self.timestamps)
        
    def _flow_2d_histogram(self, sizes, ts):
        '''
        Original Author: Tal Shapira, 
        Note: With some modifications.
        '''
        ts_norm = ((np.array(ts) - ts[0]) / (ts[-1] - ts[0])) * self.mtu
        H, xedges, yedges = np.histogram2d(sizes, ts_norm, bins=(range(0, self.mtu + 1, 1), range(0, self.mtu + 1, 1)))

        return H.astype(np.uint16)
        
        
    



#!/usr/bin/env python3
'''
CLI module
'''
from extractor import Extractor
import click


@click.command()
@click.option('--input', type=click.Path(exists=True, readable=True), help='Input PCAP file.')
@click.option('--output', type=click.Path(writable=True), default='./extracted_features', help='Output directory of extracted features.')
def entry(input, output):
    print(
        """ Parameters received: 
        Input PCAP file: {}
        Output directory: {}
        """.format(input, output))
    
    print('Starting to extract features...')
    
    Extractor(input, output).extract()
    

if __name__ == '__main__':
    entry()


from nfstream import NFStreamer, NFPlugin
import numpy as np
from scapy.all import IP, IPv6, rdpcap, raw


class NBytesPerPacket(NFPlugin):
    '''
        Extracts the first n_bytes from each packet in the flow, the bytes are taken
        from the transport layer payload (L4). if the flow have less than n_bytes bytes,
        then the rest of the bytes are zero-valued.
        
        remove_empty_payload flag tells the plugin to do not add empty payload packets such as acks in TCP.
        max_packets param determines the highest amount of packets to save/extract from flow.
    '''
    def __init__(self, n=1024, remove_empty_payload=True, max_packets=None):
        self.n = n
        self.remove_empty_payload = remove_empty_payload
        self.max_packets = max_packets
    
    def on_init(self, packet, flow):
        flow.udps.n_bytes_value = self.n
        flow.udps.n_bytes_per_packet = np.zeros((self.max_packets,self.n))
        flow.udps.n_bytes_curr_packets = 0
        
        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        if packet.payload_size == 0 and self.remove_empty_payload == True:
            return
        
        if self.max_packets is not None and flow.udps.n_bytes_curr_packets == self.max_packets:
            return
        
        amount_to_copy = min(self.n, packet.payload_size)
        if amount_to_copy == 0:
            flow.udps.n_bytes_curr_packets += 1
            return
        
        max_index_to_copy = -packet.payload_size+amount_to_copy if -packet.payload_size+amount_to_copy != 0 else None
        flow.udps.n_bytes_per_packet[flow.udps.n_bytes_curr_packets,:amount_to_copy] = np.frombuffer(self.get_payload_as_binary(packet, flow.ip_version)[-packet.payload_size:max_index_to_copy], dtype=np.uint8)
        flow.udps.n_bytes_curr_packets += 1
        
        
    def on_expire(self, flow):
        flow.udps.n_bytes_per_packet /= 255
        flow.udps.n_bytes_per_packet = [list(i) for i in flow.udps.n_bytes_per_packet]
        #[int(i) for i in list(flow.udps.n_bytes_per_packet)]
        #flow.udps.n_bytes_per_packet = list(flow.udps.n_bytes_per_packet)

    def get_payload_as_binary(self, packet, ip_version):
        if ip_version == 4:
            scapy_packet = IP(packet.ip_packet)
        elif ip_version == 6:
            scapy_packet = IPv6(packet.ip_packet)
        
        return raw(scapy_packet.payload.payload)
        


from nfstream import NFStreamer, NFPlugin
import numpy as np
from scapy.all import IP, TCP

class ProtocolHeaderFields(NFPlugin):
    '''
    wip
    '''
    def __init__(self, n_packets=32):
        self.n_packets = n_packets
    
    def on_init(self, packet, flow):
        flow.udps.protocol_header_fields_n_packets = self.n_packets
        flow.udps.protocol_header_fields = np.zeros((self.n_packets,4), dtype=np.uint32)
        
        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        if flow.bidirectional_packets <= self.n_packets and packet.protocol in [6, 17]: # TCP or UDP only.
            flow.udps.protocol_header_fields[flow.bidirectional_packets-1] = [
                packet.direction,
                packet.payload_size,
                packet.delta_time,   
                IP(packet.ip_packet)[TCP].window if packet.protocol==6 else 0
            ]

    def on_expire(self, flow):
        matrix = []
        for row in flow.udps.protocol_header_fields:
            matrix.append(list(row))
            
        flow.udps.protocol_header_fields = matrix




# used for byte frequency analysis

from nfstream import NFPlugin
import numpy as np # for bytes distribution

class ByteFrequency(NFPlugin):
    '''
        NOTE: DEPRECTATED IN FAVOR OF NPacketsByteFrequency,
        has the same behavior as NPacketsByteFrequency(n_first_packets=0)
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def on_init(self, packet, flow):
        '''
        on_init(self, packet, flow): Method called at flow creation.
        '''

        flow.udps.bytes_frequency = np.zeros(256) # of ip payload onwards
        
        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        self._add_payload_bytes_frequency(packet.ip_packet, flow.udps.bytes_frequency)

    def on_expire(self, flow):
        pass

    def _add_payload_bytes_frequency(self, payload, container):
        payload_bytes_array = np.frombuffer(payload, dtype='B') # 'B' unsigned byte
        container[payload_bytes_array] += 1


# used for byte frequency analysis

from stats.stats import IterableStats, WeightedIterableStats
from nfstream import NFPlugin
import numpy as np # for bytes distribution

class NPacketsByteFrequency(NFPlugin):
    def __init__(self, n_first_packets = 0, **kwargs):
        '''
        Params:
            n_first_packets: the byte distribution will be taken
            from the n first packets only (bidirectional), if n_first_packets=0 then
            the byte distribution will taken from all packets in the flow.
        '''
        super().__init__(**kwargs)
        self.n_first_packets = n_first_packets

    def on_init(self, packet, flow):
        '''
        on_init(self, packet, flow): Method called at flow creation.
        '''
        
        flow.udps.n_packets_byte_frequency_value = self.n_first_packets
        flow.udps.bidirectional_n_packets_byte_frequency = np.zeros(256) # of ip payload onwards
        flow.udps.src2dst_n_packets_byte_frequency = np.zeros(256) # of ip payload onwards
        flow.udps.dst2src_n_packets_byte_frequency = np.zeros(256) # of ip payload onwards
      
        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        if self.n_first_packets == 0 or flow.bidirectional_packets <= self.n_first_packets:
            if packet.direction == 0: # src -> dst
                self._add_payload_bytes_frequency(packet.ip_packet, flow.udps.src2dst_n_packets_byte_frequency)
            else: # dst -> src
                self._add_payload_bytes_frequency(packet.ip_packet, flow.udps.dst2src_n_packets_byte_frequency)
            
            # compute/update bidirectional.
            self._add_payload_bytes_frequency(packet.ip_packet, flow.udps.bidirectional_n_packets_byte_frequency)

    def on_expire(self, flow):
        '''
        Todo: zero and nan checks.
        '''
        # Normalize frequencies to get distribution
        bidirectional_n_packets_byte_distribution = flow.udps.bidirectional_n_packets_byte_frequency # / np.sum(flow.udps.bidirectional_n_packets_byte_frequency)
        src2dst_n_packets_byte_distribution       = flow.udps.src2dst_n_packets_byte_frequency # / np.sum(flow.udps.src2dst_n_packets_byte_frequency)
        dst2src_n_packets_byte_distribution       = flow.udps.dst2src_n_packets_byte_frequency # / np.sum(flow.udps.dst2src_n_packets_byte_frequency)
        # Compute statistical features
        # bidirectional
        bidirectional_stats = WeightedIterableStats(np.arange(0,256), bidirectional_n_packets_byte_distribution)
        flow.udps.bidirectional_mean_n_packets_byte_distribution             = bidirectional_stats.average()
        flow.udps.bidirectional_stdev_n_packets_byte_distribution            = bidirectional_stats.std_deviation()
        flow.udps.bidirectional_median_n_packets_byte_distribution           = bidirectional_stats.median()
        flow.udps.bidirectional_variance_n_packets_byte_distribution         = bidirectional_stats.variance()
        flow.udps.bidirectional_coeff_of_var_n_packets_byte_distribution     = bidirectional_stats.coeff_of_variation()
        flow.udps.bidirectional_skew_from_median_n_packets_byte_distribution = bidirectional_stats.skew_from_median()
        # src -> dst
        src2dst_stats = WeightedIterableStats(np.arange(0,256), src2dst_n_packets_byte_distribution)
        flow.udps.src2dst_mean_n_packets_byte_distribution             = src2dst_stats.average()
        flow.udps.src2dst_stdev_n_packets_byte_distribution            = src2dst_stats.std_deviation()
        flow.udps.src2dst_median_n_packets_byte_distribution           = src2dst_stats.median()
        flow.udps.src2dst_variance_n_packets_byte_distribution         = src2dst_stats.variance()
        flow.udps.src2dst_coeff_of_var_n_packets_byte_distribution     = src2dst_stats.coeff_of_variation()
        flow.udps.src2dst_skew_from_median_n_packets_byte_distribution = src2dst_stats.skew_from_median()
        # dst -> src
        dst2src_stats = WeightedIterableStats(np.arange(0,256), dst2src_n_packets_byte_distribution)
        flow.udps.dst2src_mean_n_packets_byte_distribution             = dst2src_stats.average()
        flow.udps.dst2src_stdev_n_packets_byte_distribution            = dst2src_stats.std_deviation()
        flow.udps.dst2src_median_n_packets_byte_distribution           = dst2src_stats.median()
        flow.udps.dst2src_variance_n_packets_byte_distribution         = dst2src_stats.variance()
        flow.udps.dst2src_coeff_of_var_n_packets_byte_distribution     = dst2src_stats.coeff_of_variation()
        flow.udps.dst2src_skew_from_median_n_packets_byte_distribution = dst2src_stats.skew_from_median()
        
        flow.udps.bidirectional_n_packets_byte_frequency = flow.udps.bidirectional_n_packets_byte_frequency.tolist() 
        flow.udps.src2dst_n_packets_byte_frequency = flow.udps.src2dst_n_packets_byte_frequency.tolist()
        flow.udps.dst2src_n_packets_byte_frequency = flow.udps.dst2src_n_packets_byte_frequency.tolist()
        # Cleanup
        

    def _add_payload_bytes_frequency(self, payload, container):
        payload_bytes_array = np.frombuffer(payload, dtype='B') # 'B' unsigned byte
        container[payload_bytes_array] += 1
